#ifndef CAMERA_H
#define CAMERA_H

typedef enum {
	PERSPECTIVE_CAMERA,
	ORTHOGRAPHIC_CAMERA
} CameraType;

typedef struct {
	float	fovy;
	float	near;
	float	far;
} PerspCamera;

typedef struct {
	float	left;
	float	right;
	float	bottom;
	float	top;
	float	near;
	float	far;
} OrthoCamera;

typedef struct {
	CameraType type;
	union {
		PerspCamera persp;
		OrthoCamera ortho;
	} cam;
} Camera;

void init_Camera( Camera* self, CameraType type );
void destroy_Camera( Camera* self );

CameraType getType_Camera( Camera const* self );
PerspCamera* getPersp_Camera( Camera* self );
OrthoCamera* getOrtho_Camera( Camera* self );

void setParams_PerspCamera( PerspCamera* self, float fovy, float near, float far );
float getFOVY_PerspCamera( PerspCamera const* self );
float getNear_PerspCamera( PerspCamera const* self );
float getFar_PerspCamera( PerspCamera const* self );

void setParams_OrthoCamera( OrthoCamera* self, float left, float right,
	float bottom, float top, float near, float far );
float getLeft_OrthoCamera( OrthoCamera const* self );
float getRight_OrthoCamera( OrthoCamera const* self );
float getBottom_OrthoCamera( OrthoCamera const* self );
float getTop_OrthoCamera( OrthoCamera const* self );
float getNear_OrthoCamera( OrthoCamera const* self );
float getFar_OrthoCamera( OrthoCamera const* self );

#endif	/* CAMERA_H */

