#ifndef NODE_H
#define NODE_H

#include <containers/stack.h>
#include <containers/vec_.h>
#include <types.h>

typedef enum NodeType {
	NODE_RAW,
	NODE_SCENE,
	NODE_MESH,
	NODE_CAMERA,
	NODE_OBJECT2D,
	NODE_BILLBOARD,
} NodeType;

typedef struct Node {
	NodeType	type;
	struct Node*	parent;
	Vec		children;
} Node;

typedef struct {
	Stack	stack;
	Node*	node;
	Node*	next;
	uint_t	i;
} NodeIterator;

/* interface */
void nodeConstruct(Node* node);
void nodeDestroy(Node* node);
void nodeAttach(Node* node, Node* child);
uint_t nodeGetNumChildren(const Node* node);
Node* nodeGetParent(Node* node);
Node* nodeGetChild(Node* node, uint_t i);

void nodeIteratorConstruct(NodeIterator* iterator);
Node* nodeIteratorInc(NodeIterator* iterator);
void nodeIteratorReset(NodeIterator* iterator, Node* node);

#endif	/* NODE_H */

