#include <containers/hash_.h>
#include <abort.h>
#include <malloc.h>

Hash_* hash_New(size_t keysize, size_t valuesize, keys_equ_f keys_equ) {
	Hash_* hash = malloc(sizeof(Hash_));
	if(NULL == hash)
		abort();
	hash_Construct(hash, keysize, valuesize, keys_equ);
	return hash;
}

void hash_Delete(Hash_* hash) {

	hash_Destroy(hash);
	free(hash);
}

void hash_Construct(Hash_* hash, size_t keysize, size_t valuesize, keys_equ_f keys_equ) {

	hash->keysize = keysize;
	hash->valuesize = valuesize;
	hash->keys_equ = keys_equ;
	init_Vec(&hash->pairs, sizeof(Pair_), 0);
}

void hash_Destroy(Hash_* hash) {

	for(uint32_t i=0; i<hash_GetNumKeys(hash); i++) {
		pair_Delete(hash_GetPair(hash, i));
	}
	destroy_Vec(&hash->pairs);
}

void hash_Insert(Hash_* hash, void* key, void* value) {

	Pair_* pair = NULL;
	for(uint32_t i=0, n=getLength_Vec(&hash->pairs); i<n; i++) {
		pair = getElement_Vec(&hash->pairs, i);
		if(hash->keys_equ(key, pair->key))
			break;
		pair = NULL;
	}

	if(NULL == pair) {
		/* didn't find desired key - insert it */
		Pair_* newpair = pair_New(hash->keysize, hash->valuesize, key);
		pushBack_Vec(&hash->pairs, newpair);
		pair = newpair;
	}

	pushBack_Vec(&pair->vals, value);
}

inline uint32_t hash_GetNumKeys(Hash_* hash) {
	return getLength_Vec(&hash->pairs);
}

inline Pair_* hash_GetPair(Hash_* hash, uint32_t i) {
	return getElement_Vec(&hash->pairs, i);
}

