#include <abort.h>
#include <malloc.h>
#include <stream.h>
#include <ultra64.h>
#include <sys/region.h>
#include <math/vector.h>
#include <math/math.h>
#include <sys/sys.h>
#include <vertex.h>
#include <node/lux.h>
#include <node/xformer.h>
#include <types.h>
#include <dltrans.h>
#include <rectangle.h>

#define JUMP_NATIVE_DL_SEGMENT	0x8
#define VTXBUFFER_MAXELEMENTS	64
#define IDXBUFFER_MAXELEMENTS	64
#define MAX_LIGHTS 				8
#define MAX_NODES 				64
#define	MAX_MATERIALS 			128
#define MAX_TRISOURCES			256

#define G_CC_MYMODE				0, 0, 0, SHADE, 0, 0, 0, PRIMITIVE
#define MY_G_CC_DECAL_XLU		0, 0, 0, TEXEL0, 0, 0, 0, PRIMITIVE
//#define MY_G_CC_DECAL_XLU		0, 0, 0, TEXEL0, 0, 0, 0, TEXEL0

static Light rsp_light[MAX_MATERIALS][MAX_LIGHTS];
static Lights0 rsp_ambient_light = gdSPDefLights0( 0x00, 0x00, 0x40 );

static Material material[MAX_MATERIALS];
static Lux const* lux[MAX_LIGHTS];
static Vector3f const* lux_pos[MAX_LIGHTS];
static Matrix* cpu_mtx[MAX_NODES];
static LookAt lookat[MAX_NODES];
static Triangle const* tri_src[MAX_TRISOURCES];

static Vp vp = {
	{
		//{ (BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ/2, 0 },			// G_MAXZ = 0x03FF
		//{ (BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ/2, 0 }
		{ (BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ/2, 0 },
		{ (BUFFER_W/2)*4, (BUFFER_H/2)*4, G_MAXZ/2, 0 }
	}
};

static Gfx rspinit_dl[] = {
	gsSPViewport( &vp ),
	//gsSPSetGeometryMode( G_CULL_FRONT ),
	//gsSPClearGeometryMode( G_CULL_BOTH ),
	gsSPClipRatio( FRUSTRATIO_1 ),
	gsSPEndDisplayList()
};

static Gfx rdpinit_dl[] = {
	gsDPSetCycleType( G_CYC_1CYCLE ),
	gsDPPipelineMode( G_PM_NPRIMITIVE ),
	gsDPSetScissor( G_SC_NON_INTERLACE, 0, 0, BUFFER_W, BUFFER_H ),
	gsDPSetTextureLOD( G_TL_TILE ),
	gsDPSetTextureLUT( G_TT_NONE ),
	gsDPSetTextureDetail( G_TD_CLAMP ),
	gsDPSetTexturePersp( G_TP_PERSP ),
	gsDPSetTextureFilter( G_TF_BILERP ),
	gsDPSetTextureConvert( G_TC_FILT ),
	gsDPSetCombineMode( G_CC_SHADE, G_CC_SHADE ),
	gsDPSetCombineKey( G_CK_NONE ),
	gsDPSetAlphaCompare( G_AC_NONE ),
	gsDPSetRenderMode( G_RM_OPA_SURF, G_RM_OPA_SURF2 ),
	//gsDPSetRenderMode( G_RM_XLU_SURF, G_RM_XLU_SURF2 ),
	//gsDPSetBlendColor( 0xFF, 0xFF, 0x00, 0x20 ),
	gsDPSetColorDither( G_CD_DISABLE ),
	gsDPPipeSync(),
	gsSPEndDisplayList()
};

void init_DLTrans( DLTrans* self ) {
	init_Vec( &self->rspMatrices, sizeof(Mtx), ALIGN_8B );
	resize_Vec( &self->rspMatrices, MAX_NODES );

	init_Vec( &self->vtxBuffer, sizeof(Vtx), ALIGN_8B );
	//resize_Vec( &self->vtxBuffer, VTXBUFFER_MAXELEMENTS );
	self->scale = 1.0f;
	setColor_Material( &self->material, 1.0f, 1.0f, 1.0f, 1.0f );
}

void destroy_DLTrans( DLTrans* self ) {
	destroy_Vec( &self->vtxBuffer );
}

void process_DLTrans( DLTrans* self, DList const* dl_in, NativeDL* dl_out ) {
	reset_Vec( &self->rspMatrices );
	reset_Vec( &self->vtxBuffer );

	NativeDL* rsp_dl = dl_out;
	uint16_t current_mtx_i = 0;
	uint16_t current_lux_i = 0;
	uint16_t current_mtl_i = 0;
	uint16_t current_lookat_i = 0;
	uint16_t current_tri_src = 0;
	uint32_t seg0_adr = 0;

	Stream stream;
	init_Stream( &stream,
		getData_DList( dl_in ),
		getSize_DList( dl_in )
	);

	name_t name;
	do {
		name = *( name_t* )getData_Stream( &stream );
		switch( name ) {
			case Start: {
				(void*)extractElement_Stream( &stream,
					sizeof( Command_Start )
				);

				addCommand_NativeDL( rsp_dl,
					gsSPSegment( 0x0, 0x0 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPSegment( JUMP_NATIVE_DL_SEGMENT, 0x0 )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( rspinit_dl )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( rdpinit_dl )
				);
				break;
			}
			case SetVertexSource: {
				Command_SetVertexSource const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetVertexSource )
					);
				self->vtxFormat = cmd->format;
				self->vtxSource = cmd->data;
				break;
			}
			case SetTriangleSource: {
				Command_SetTriangleSource const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetTriangleSource )
					);
				tri_src[current_tri_src++] = cmd->data;
				break;
			}
			case SetVertexScale: {
				Command_SetVertexScale const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetVertexScale )
					);
				self->scale = cmd->scale;
				break;
			}
			case LoadVertexBuffer: {
				Command_LoadVertexBuffer const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadVertexBuffer )
					);

				uint32_t n = cmd->nvts;
				if( n + cmd->dst_ofs > VTXBUFFER_MAXELEMENTS ) {
					//abort();
				}

				Vertex_generic* vtxSource = ( Vertex_generic* )self->vtxSource + cmd->src_ofs;
				while( n-- ) {
					Vtx vtx;
					vtx.v.ob[0] = vtxSource->pos.x * self->scale;
					vtx.v.ob[1] = vtxSource->pos.y * self->scale;
					vtx.v.ob[2] = vtxSource->pos.z * self->scale;
					vtx.v.tc[0] = vtxSource->texcoords.x * 32 * getWidth_Texture( self->tex );
					vtx.v.tc[1] = vtxSource->texcoords.y * 32 * getHeight_Texture( self->tex );
					vtx.v.cn[0] = vtxSource->normal.x * 127;
					vtx.v.cn[1] = vtxSource->normal.y * 127;
					vtx.v.cn[2] = vtxSource->normal.z * 127;
					vtx.v.cn[3] = 0x7F;
					pushBack_Vec( &self->vtxBuffer, &vtx );

					vtxSource++;
				}

				break;
			}
			case LoadIndexBuffer: {
				abort();
				/*
				Command_LoadIndexBuffer const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadIndexBuffer )
					);

				uint32_t n = cmd->nids;
				if( n > IDXBUFFER_MAXELEMENTS ) {
					abort();
				}

				uint32_t* idxSource = self->triSource->ids + cmd->src_ofs;
				uint32_t* idx = self->idxBuffer + cmd->dst_ofs;
				while( n-- ) {
					*idx++ = *idxSource++;
				}
				*/
				break;
			}
			case DrawPrimitives: {
				Command_DrawPrimitives const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_DrawPrimitives )
					);
				if( current_tri_src >= MAX_TRISOURCES ) {
					abort();
				}

				Vtx* vtx = getElement_Vec( &self->vtxBuffer, cmd->ofs );
				Triangle const* tri = tri_src[current_tri_src - 1];	//self->triSource;
				if( PrimTriangle == cmd->type ) {
					for( uint32_t i=0; i<cmd->num; ++i, ++tri ) {
						addCommand_NativeDL( rsp_dl,
							gsSPVertex( &vtx[ tri->ids[0] ], 1, 0 )
						);
						addCommand_NativeDL( rsp_dl,
							gsSPVertex( &vtx[ tri->ids[1] ], 1, 1 )
						);
						addCommand_NativeDL( rsp_dl,
							gsSPVertex( &vtx[ tri->ids[2] ], 1, 2 )
						);
						addCommand_NativeDL( rsp_dl,
							gsSP1Triangle( 0, 1, 2, 0 )
						);
					}
				}
				else {
					abort();
				}

				break;
			}
			case SetBuffer: {
				Command_SetBuffer const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetBuffer )
					);
				Surface* surface;

				if( ColorBuffer == cmd->buftype ) {
					surface = &self->surfColor;
					addCommand_NativeDL( rsp_dl,
						gsDPSetColorImage(
							G_IM_FMT_RGBA,
							G_IM_SIZ_16b,
							cmd->width,
							cmd->buf
						)
					);
				} else if( ZBuffer == cmd->buftype ) {
					surface = &self->surfZ;
					addCommand_NativeDL( rsp_dl,
						gsDPSetDepthImage( cmd->buf )
					);
				} else {
					abort();
				}

				surface->pixfmt = cmd->format;
				surface->data = cmd->buf;
				surface->width = cmd->width;
				surface->height = cmd->height;

				break;
			}
			case ClearBuffers: {
				Command_ClearBuffers const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_ClearBuffers )
					);

				if( cmd->types & ColorBuffer ) {
					uint32_t clrval = GPACK_RGBA5551(
						( uint8_t )( 255 * self->clearColor.x ),
						( uint8_t )( 255 * self->clearColor.y ),
						( uint8_t )( 255 * self->clearColor.z ),
						1
					);
					clrval = clrval << 16 | clrval;
					Surface* surface = &self->surfColor;

					addCommand_NativeDL( rsp_dl,
						gsDPSetColorImage(
							G_IM_FMT_RGBA,
							G_IM_SIZ_16b,
							surface->width,
							surface->data
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetCycleType( G_CYC_FILL )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_NOOP,
							G_RM_NOOP2
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetFillColor( clrval )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPFillRectangle(
							0,
							0,
							surface->width - 1,
							surface->height - 1
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPPipeSync()
					);
				}
				if( cmd->types & ZBuffer ) {
					uint32_t clrval = GPACK_ZDZ(
						( uint16_t )( G_MAXFBZ * self->clearZVal ),
						0
					);
					clrval = clrval << 16 | clrval;
					Surface* surface = &self->surfZ;

					addCommand_NativeDL( rsp_dl,
						gsDPSetColorImage(
							G_IM_FMT_RGBA,
							G_IM_SIZ_16b,
							surface->width,
							surface->data
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetCycleType( G_CYC_FILL )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_NOOP,
							G_RM_NOOP2
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetFillColor( clrval )
					);
					addCommand_NativeDL( rsp_dl,
						gsDPFillRectangle(
							0,
							0,
							surface->width - 1,
							surface->height - 1
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPPipeSync()
					);
				}

				break;
			}
			case SetClearColor: {
				Command_SetClearColor const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetClearColor )
					);

				init_Vector4f( &self->clearColor, cmd->r, cmd->g, cmd->b, cmd->a );

				break;
			}
			case SetClearZValue: {
				Command_SetClearZValue const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetClearZValue )
					);

				self->clearZVal = cmd->val;

				break;
			}
			case LoadTexture: {
				Command_LoadTexture const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadTexture )
					);

				self->tex = cmd->texture;
				Texture const* texture = cmd->texture;
				Rectangle const* tile = &cmd->tile;
				bool const prevent_seams = cmd->prevent_seams;

				if( TEXTURE_FMT_RGBA32 == getFormat_Texture( texture ) ) {
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTextureTile(
							getPixelsConst_Texture( texture ),
							G_IM_FMT_RGBA,
							G_IM_SIZ_32b,
							getWidth_Texture( texture ),
							getHeight_Texture( texture ),
							
							tile->p0x,
							tile->p0y,
							tile->p1x,
							tile->p1y,

							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
							log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
						/*
						gsDPLoadTextureBlock(
							texture->pixels,
							G_IM_FMT_RGBA,
							G_IM_SIZ_32b,
							texture->width,
							texture->height,
							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2uint( texture->width ),
							log2uint( texture->height ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
						*/
					);
					addCommand_NativeDL( rsp_dl, gsDPSetTextureLUT( G_TT_NONE ) );
				} else if( TEXTURE_FMT_RGBA16 == getFormat_Texture( texture ) ) {
					bool s_wrap = false;
					if( tile->p1x >= getWidth_Texture( texture ) ) {
						//s_wrap = true;
					}

					bool t_wrap = false;
					if( tile->p1y >= getHeight_Texture( texture ) ) {
						//t_wrap = true;
					}

					if( false == t_wrap ) {
						addCommand_NativeDL( rsp_dl,
							gsDPLoadTextureTile(
								(uint16_t const*)getPixelsConst_Texture( texture ) - getWidth_Texture( texture ) * (true == s_wrap),
								G_IM_FMT_RGBA,
								G_IM_SIZ_16b,
								getWidth_Texture( texture ),
								getHeight_Texture( texture ),

								tile->p0x,
								tile->p0y,
								tile->p1x + prevent_seams,
								tile->p1y + prevent_seams + (true == s_wrap),

								0,
								G_TX_WRAP,
								G_TX_WRAP,
								log2_has_frac( prevent_seams + 1 + tile->p1x - tile->p0x )
									+ log2uint( prevent_seams + 1 + tile->p1x - tile->p0x ),
								log2_has_frac( prevent_seams + 1 + tile->p1y - tile->p0y )
									+ log2uint( prevent_seams + 1 + tile->p1y - tile->p0y ),
								G_TX_NOLOD,
								G_TX_NOLOD
							)
						);
					} else {	// true == t_wrap
						addCommand_NativeDL( rsp_dl,
							gsDPLoadMultiTile(
								(uint16_t const*)getPixelsConst_Texture( texture ) - getWidth_Texture( texture ) * (true == s_wrap),
								((((tile->p1x - tile->p0x + 1) * G_IM_SIZ_16b_TILE_BYTES) + 7) >> 3)
									* (texture->height - tile->p0y + (true == s_wrap)),
								0,	// rtile
								G_IM_FMT_RGBA,
								G_IM_SIZ_16b,
								getWidth_Texture( texture ),
								getHeight_Texture( texture ),

								tile->p0x,
								0,
								tile->p1x,
								tile->p1y - getHeight_Texture( texture ) + (true == s_wrap),

								0,
								G_TX_WRAP,
								G_TX_WRAP,
								log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
								log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
								G_TX_NOLOD,
								G_TX_NOLOD
							)
						);
						addCommand_NativeDL( rsp_dl,
							gsDPLoadTextureTile(
								(uint16_t const*)getPixelsConst_Texture( texture ) - getWidth_Texture( texture ) * (true == s_wrap),
								G_IM_FMT_RGBA,
								G_IM_SIZ_16b,
								getWidth_Texture( texture ),
								getHeight_Texture( texture ),

								tile->p0x,
								tile->p0y,
								tile->p1x,
								getHeight_Texture( texture ) - 1 + (true == s_wrap),

								0,
								G_TX_WRAP,
								G_TX_WRAP,
								log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
								log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
								G_TX_NOLOD,
								G_TX_NOLOD
							)
						);
						/*
						addCommand_NativeDL( rsp_dl,
							gsDPSetTileSize(
								G_TX_RENDERTILE,
								tile->p0x << G_TEXTURE_IMAGE_FRAC,
								tile->p0y << G_TEXTURE_IMAGE_FRAC,
								tile->p1x << G_TEXTURE_IMAGE_FRAC,
								tile->p1y << G_TEXTURE_IMAGE_FRAC
							)
						);
						*/
					}

					if( true == s_wrap ) {
						addCommand_NativeDL( rsp_dl,
							gsDPPipeSync()
						);
						addCommand_NativeDL( rsp_dl,
							gsDPSetTile( G_IM_FMT_RGBA, G_IM_SIZ_16b,
								((((tile->p1x - tile->p0x + 1) * G_IM_SIZ_16b_LINE_BYTES) + 7) >> 3),	// line
								//((((tile->p1x - tile->p0x + 1) * G_IM_SIZ_16b_LINE_BYTES) + 7) >> 3),	// tmem
								0, //(((((texture->width - tile->p0x + 1) * G_IM_SIZ_16b_TILE_BYTES) + 7) >> 3)-0),	// tmem
								G_TX_RENDERTILE, 0,
								G_TX_WRAP,
								log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
								G_TX_NOLOD,
								G_TX_WRAP,
								log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
								G_TX_NOLOD
							)
						);
					}
#if 1
					addCommand_NativeDL( rsp_dl,
						gsDPSetTileSize(
							G_TX_RENDERTILE,
							(tile->p0x - prevent_seams) << G_TEXTURE_IMAGE_FRAC,
							(tile->p0y - prevent_seams) << G_TEXTURE_IMAGE_FRAC,
							(tile->p1x + 0) << G_TEXTURE_IMAGE_FRAC,
							(tile->p1y + 0) << G_TEXTURE_IMAGE_FRAC
						)
					);
#endif
					addCommand_NativeDL( rsp_dl,
						gsDPSetTextureLUT( G_TT_NONE )
					);

				} else if( TEXTURE_FMT_CI4 == getFormat_Texture( texture ) ) {
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTextureTile_4b(
							getPixelsConst_Texture( texture ),
							G_IM_FMT_CI,
							getWidth_Texture( texture ),
							getHeight_Texture( texture ),
							
							tile->p0x,
							tile->p0y,
							tile->p1x,
							tile->p1y,

							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2_has_frac( 1 + tile->p1x - tile->p0x ) + log2uint( 1 + tile->p1x - tile->p0x ),
							log2_has_frac( 1 + tile->p1y - tile->p0y ) + log2uint( 1 + tile->p1y - tile->p0y ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTLUT_pal16(
							0,
							getPaletteConst_Texture( texture )
						)
					);
					addCommand_NativeDL( rsp_dl,
						gsDPSetTextureLUT( G_TT_RGBA16 )
					);
				} else {
					addCommand_NativeDL( rsp_dl,
						gsDPLoadTextureBlock(
							getPixelsConst_Texture( texture ),
							G_IM_FMT_RGBA,
							G_IM_SIZ_32b,
							getWidth_Texture( texture ),
							getHeight_Texture( texture ),
							0,
							G_TX_WRAP,
							G_TX_WRAP,
							log2uint( getWidth_Texture( texture ) ),
							log2uint( getHeight_Texture( texture ) ),
							G_TX_NOLOD,
							G_TX_NOLOD
						)
					);
					addCommand_NativeDL( rsp_dl, gsDPSetTextureLUT( G_TT_NONE ) );
					//abort( );
				}
				break;
			}
			case Jump: {	// beware: allocation on every invocation!
				abort();
				Command_Jump const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_Jump )
					);
				NativeDL* child_dl = new_NativeDL();

				process_DLTrans( self, cmd->dl, child_dl );
				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( child_dl )
				);

				break;
			}
			case End: {
				(void*)extractElement_Stream( &stream,
					sizeof( Command_End )
				);

				addCommand_NativeDL( rsp_dl,
					gsSPEndDisplayList()
				);
				break;
			}
			case SetTextureScale: {
				Command_SetTextureScale const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetTextureScale )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPTexture(
						(s32)( 0x8000 * cmd->s ),
						(s32)( 0x8000 * cmd->t ),
						0,
						G_TX_RENDERTILE,
						G_ON
					)
				);

				break;
			}
			case LoadMatrix: {
				Command_LoadMatrix const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadMatrix )
					);

				if( current_mtx_i >= MAX_NODES ) {
					abort();
				}

				cpu_mtx[current_mtx_i] = cmd->mtx;

				Mtx* mtx_native = getElement_Vec( &self->rspMatrices, current_mtx_i++ );
				guMtxF2L( cmd->mtx->m, mtx_native );

				if( MATRIX_PROJECTION == cmd->type ) {
					addCommand_NativeDL( rsp_dl,
						gsSPMatrix(
							mtx_native,
							G_MTX_PROJECTION | G_MTX_LOAD | G_MTX_NOPUSH
						)
					);
				} else if( MATRIX_MODELVIEW == cmd->type ) {
					addCommand_NativeDL( rsp_dl,
						gsSPMatrix(
							mtx_native,
							G_MTX_MODELVIEW | G_MTX_LOAD | G_MTX_NOPUSH
						)
					);
				} else {
					abort();
				}
				break;
			}
			case LoadLight: {
				Command_LoadLight const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_LoadLight )
					);

				uint16_t idx = cmd->idx;
				if( idx > MAX_LIGHTS ) {
					abort();
				}

				/* color */
				lux[current_lux_i] = cmd->light;
				lux_pos[current_lux_i] = cmd->pos;
				++current_lux_i;

				break;
			}
			case Sync: {
				(void*)extractElement_Stream( &stream,
					sizeof( Command_Sync )
				);

				addCommand_NativeDL( rsp_dl,
					gsDPFullSync()
				);
				break;
			}
			case PipeSync: {
				(void*)extractElement_Stream( &stream,
					sizeof( Command_PipeSync )
				);

				addCommand_NativeDL( rsp_dl,
					gsDPPipeSync()
				);
				break;
			}
			case JumpToTranslated: {
				Command_JumpToTranslated const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_JumpToTranslated )
					);

				uint16_t mtx_i = current_mtx_i - 1;
				if( 0 == current_mtx_i ) {
					mtx_i = 0;
				}
				uint16_t mtl_i = current_mtl_i - 1;
				if( 0 == current_mtl_i ) {
					mtl_i = 0;
				}
				
				// set material opacity
				addCommand_NativeDL( rsp_dl,
					gsDPPipeSync()
				);
				addCommand_NativeDL( rsp_dl,
					gsDPSetPrimColor(
						0, 0, 0x00, 0x00, 0x00,
						getColor_Material( &material[mtl_i] )->t * 255
					)
				);

				Vector3f center, dir;
				Matrix* mtx;
				for( uint16_t i=0; i<current_lux_i; ++i ) {
					mtx = cpu_mtx[mtx_i];
					init_Vector3f( &center,
						MATRIX_GET(*mtx, 3, 0),
						MATRIX_GET(*mtx, 3, 1),
						MATRIX_GET(*mtx, 3, 2)
					);
					sub_Vector3f( &dir, &center, lux_pos[i] );
					normalize_Vector3f( &dir, &dir );

					Light_t* light = &rsp_light[mtl_i][i].l;
					Vector4f const* mtl_color = &material[mtl_i].color;
					/* multiply lights by material colors before rendering */
					light->col[0]
						= light->colc[0]
						= 255 * lux[i]->color.x * mtl_color->x;

					light->col[1]
						= light->colc[1]
						= 255 * lux[i]->color.y * mtl_color->y;

					light->col[2]
						= light->colc[2]
						= 255 * lux[i]->color.z * mtl_color->z;

					/* direction */
					light->dir[0] = -127*dir.x;
					light->dir[1] = -127*dir.y;
					light->dir[2] = -127*dir.z;

					addCommand_NativeDL( rsp_dl,
						gsSPLight( &rsp_light[mtl_i][i], i + 1 )
					);
				}
				addCommand_NativeDL( rsp_dl,
					gsSPLight( &rsp_ambient_light, current_lux_i + 1 )
				);

				s32 num_lights;
				switch( current_lux_i ) {
					case 0:
						num_lights = NUMLIGHTS_0;
						break;
					case 1:
						num_lights = NUMLIGHTS_1;
						break;
					case 2:
						num_lights = NUMLIGHTS_2;
						break;
					case 3:
						num_lights = NUMLIGHTS_3;
						break;
					case 4:
						num_lights = NUMLIGHTS_4;
						break;
					case 5:
						num_lights = NUMLIGHTS_5;
						break;
					case 6:
						num_lights = NUMLIGHTS_6;
						break;
					case 7:
						num_lights = NUMLIGHTS_7;
						break;
					default:
						num_lights = NUMLIGHTS_7;
						break;
				}
				addCommand_NativeDL( rsp_dl,
					gsSPNumLights( num_lights )
				);

				addCommand_NativeDL( rsp_dl,
					gsSPSegment( 0x0, seg0_adr )
				);
				addCommand_NativeDL( rsp_dl,
					gsSPDisplayList( (Gfx*)((uintptr_t)(cmd->dl) | (JUMP_NATIVE_DL_SEGMENT << 24)) )
				);

				addCommand_NativeDL( rsp_dl,
					gsSPSegment( 0x0, 0x0 )
				);
				break;
			}
			case SetCombineMode: {
				Command_SetCombineMode const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetCombineMode )
					);

				if( (CM_SHADE == cmd->mode) && (CMA_OPAQUE == cmd->amode)  ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetCombineMode(
							G_CC_SHADE,
							G_CC_SHADE
						)
					);
				} else if( (CM_SHADE == cmd->mode) && (CMA_TRANSLUCENT == cmd->amode) ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetCombineMode(
							G_CC_MYMODE,
							G_CC_MYMODE
						)
					);
				} else if( (CM_DECAL == cmd->mode ) && (CMA_OPAQUE == cmd->amode) ) {
#define MY_G_CC_DECALRGBA	0, 0, 0, TEXEL0, 0, 0, 0, TEXEL0
					addCommand_NativeDL( rsp_dl,
						gsDPSetCombineMode(
							MY_G_CC_DECALRGBA,
							MY_G_CC_DECALRGBA
						)
					);
				} else if( (CM_DECAL == cmd->mode ) && (CMA_TRANSLUCENT == cmd->amode) ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetCombineMode(
							MY_G_CC_DECAL_XLU,
							MY_G_CC_DECAL_XLU
						)
					);
				} else {
					abort( );
				}
				break;
			}
			case SetRenderMode: {
				Command_SetRenderMode const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetRenderMode )
					);

				addCommand_NativeDL( rsp_dl,			// !!! maybe user should explicitly request pipe sync?
					gsDPPipeSync()
				);
				addCommand_NativeDL( rsp_dl,
					gsDPSetCycleType( G_CYC_1CYCLE )
				);

				if( RM_ZB_OPAQUE == cmd->mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_ZB_OPA_SURF,
							G_RM_ZB_OPA_SURF2
						)
					);
				} else if( RM_OPAQUE == cmd->mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_OPA_SURF,
							G_RM_OPA_SURF2
						)
					);
				} else if( RM_ZB_TRANSLUCENT == cmd->mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_ZB_XLU_SURF,
							G_RM_ZB_XLU_SURF2
						)
					);
				} else if( RM_ZB_TEX_EDGE == cmd->mode ) {
					addCommand_NativeDL( rsp_dl,
						gsDPSetRenderMode(
							G_RM_AA_ZB_TEX_EDGE,
							G_RM_AA_ZB_TEX_EDGE2
						)
					);
				} else {
					abort();
				}

				// FOG
#if 0
				addCommand_NativeDL( rsp_dl,
					gsDPSetCycleType( G_CYC_2CYCLE )
				);
				addCommand_NativeDL( rsp_dl,
					gsDPSetRenderMode(
						G_RM_FOG_SHADE_A,
						G_RM_AA_ZB_TEX_EDGE2
					)
				);
				addCommand_NativeDL( rsp_dl,
					gsSPFogPosition(
						990,
						1000
					)
				);
				addCommand_NativeDL( rsp_dl,
					gsSPSetGeometryMode( G_FOG )
				);
				addCommand_NativeDL( rsp_dl,
					gsDPSetFogColor( 255, 255, 255, 255 )
				);
#endif
				break;
			}
			case NormalizePerspective: {
				Command_NormalizePerspective const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_NormalizePerspective )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPPerspNormalize( (u16) cmd->scale)
				);
				break;
			}
			case SetLookAt: {
				Command_SetLookAt const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetLookAt )
					);

				static float mreflect[4][4];
				guLookAtReflectF( mreflect, &lookat[current_lookat_i],
					cmd->eye.x, cmd->eye.y, cmd->eye.z,
					cmd->object.x, cmd->object.y, cmd->object.z,
					cmd->up.x, cmd->up.y, cmd->up.z
				);

				addCommand_NativeDL( rsp_dl,
					gsSPLookAt( &lookat[current_lookat_i] )
				);

				++current_lookat_i;
				break;
			}
			case SetGeometryMode: {
				Command_SetGeometryMode const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetGeometryMode )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPSetGeometryMode( cmd->gmode )
				);
				break;
			}
			case ClearGeometryMode: {
				Command_ClearGeometryMode const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_ClearGeometryMode )
					);

				addCommand_NativeDL( rsp_dl,
					gsSPClearGeometryMode( cmd->gmode )
				);
				break;
			}
			case SetViewport: {
				Command_SetViewport const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetViewport )
					);

				vp.vp.vscale[0] = (getWidth_Viewport( cmd->vp ) / 2)*4;
				vp.vp.vscale[1] = (getHeight_Viewport( cmd->vp ) / 2)*4;
				vp.vp.vscale[2] = (getFar_Viewport( cmd->vp ) - getNear_Viewport( cmd->vp )) / 2;
				vp.vp.vscale[3] = 0;

				vp.vp.vtrans[0] = (getWidth_Viewport( cmd->vp ) / 2)*4;
				vp.vp.vtrans[1] = (getHeight_Viewport( cmd->vp ) / 2)*4;
				vp.vp.vscale[2] = (getFar_Viewport( cmd->vp ) + getNear_Viewport( cmd->vp )) / 2;
				vp.vp.vtrans[3] = 0;

				addCommand_NativeDL( rsp_dl,
					gsSPViewport( &vp )
				);
				break;
			}

			case SetMaterial: {
				Command_SetMaterial const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetMaterial )
					);

				if( current_mtl_i > MAX_MATERIALS ) {
					abort();
				}
				clone_Material( &cmd->mtl, &material[current_mtl_i] );
				++current_mtl_i;

				break;
			}

			case SetSegment: {
				Command_SetSegment const* cmd =
					extractElement_Stream( &stream,
						sizeof( Command_SetSegment )
					);

				seg0_adr = (uint32_t)cmd->adr;
				break;
			}
			
			default: {
				abort( );
				break;
			}
		}
	} while( End != name );

	destroy_Stream( &stream );
}

