#include <math/vector.h>

float sqrtf( float );

inline void init_Vector2f( Vector2f* self, float x, float y ) {
	self->x = x;
	self->y = y;
}

inline void init_Vector3f( Vector3f* self, float x, float y, float z ) {
	self->x = x;
	self->y = y;
	self->z = z;
}

inline void init_Vector4f( Vector4f* self, float x, float y, float z, float t ) {
	self->x = x;
	self->y = y;
	self->z = z;
	self->t = t;
}

inline void add_Vector3f( Vector3f* result, Vector3f const* u, Vector3f const* v ) {
	result->x = u->x + v->x;
	result->y = u->y + v->y;
	result->z = u->z + v->z;
}

inline void saturatedAdd_Vector3f( Vector3f* result, Vector3f const* u, Vector3f const* v ) {
	add_Vector3f( result, u, v );

	if( result->x > 1.0f ) {
		result->x = 1.0f;
	}
	if( result->y > 1.0f ) {
		result->y = 1.0f;
	}
	if( result->z > 1.0f ) {
		result->z = 1.0f;
	}
}

inline void sub_Vector3f( Vector3f* result, Vector3f const* u, Vector3f const* v ) {
	result->x = u->x - v->x;
	result->y = u->y - v->y;
	result->z = u->z - v->z;
}

inline void scale_Vector3f( Vector3f* result, Vector3f const* v, float scale ) {
	result->x = v->x * scale;
	result->y = v->y * scale;
	result->z = v->z * scale;
}


inline float length_Vector3f( Vector3f const* v ) {
	return sqrtf( v->x * v->x + v->y * v->y + v->z * v->z );
}

inline void normalize_Vector3f( Vector3f* result, Vector3f const* v ) {
	float l = length_Vector3f( v );
	if( 0 == l ) {
		return;
	}

	result->x = v->x / l;
	result->y = v->y / l;
	result->z = v->z / l;
}

inline void crossProduct_Vector3f( Vector3f* result, Vector3f const* u, Vector3f const* v ) {
	result->x = (u->y) * (v->z) - (u->z) * (v->y);
	result->y = (u->z) * (v->x) - (u->x) * (v->z);
	result->z = (u->x) * (v->y) - (u->y) * (v->x);
}
