#include <abort.h>
#include <string.h>
#include <node/camera.h>

void init_Camera( Camera* self, CameraType type ) {
	self->type = type;
	switch( type ) {
		case PERSPECTIVE_CAMERA: {
			setParams_PerspCamera( getPersp_Camera( self ), 63.74f, 1.33f, 1000.0f );
			break;
		}
		case ORTHOGRAPHIC_CAMERA: {
			setParams_OrthoCamera( getOrtho_Camera( self ), -160.0f, 160.0f, -120.0f, 120.0f, 1.33f, 1000.0f );
			break;
		}
		default: {
			abort();
		}
	}
}

void destroy_Camera( Camera* self ) {
	if( self ) {}
}

CameraType getType_Camera( Camera const* self ) {
	return self->type;
}

PerspCamera* getPersp_Camera( Camera* self ) {
	return &self->cam.persp;
}

OrthoCamera* getOrtho_Camera( Camera* self ) {
	return &self->cam.ortho;
}

void setParams_PerspCamera( PerspCamera* self, float fovy, float near, float far ) {
	self->fovy = fovy;
	self->near = near;
	self->far = far;
}

inline float getFOVY_PerspCamera( PerspCamera const* self ) {
	return self->fovy;
}

inline float getNear_PerspCamera( PerspCamera const* self ) {
	return self->near;
}

inline float getFar_PerspCamera( PerspCamera const* self ) {
	return self->far;
}


inline void setParams_OrthoCamera( OrthoCamera* self, float left, float right,
	float bottom, float top, float near, float far ) {

	self->left = left;
	self->right = right;
	self->bottom = bottom;
	self->top = top;
	self->near = near;
	self->far = far;
}

inline float getLeft_OrthoCamera( OrthoCamera const* self ) {
	return self->left;
}

inline float getRight_OrthoCamera( OrthoCamera const* self ) {
	return self->right;
}

inline float getBottom_OrthoCamera( OrthoCamera const* self ) {
	return self->bottom;
}

inline float getTop_OrthoCamera( OrthoCamera const* self ) {
	return self->top;
}

inline float getNear_OrthoCamera( OrthoCamera const* self ) {
	return self->bottom;
}

inline float getFar_OrthoCamera( OrthoCamera const* self ) {
	return self->far;
}

