#include <abort.h>
#include <string.h>
#include <types.h>
#include <node/tnode.h>

void tnodeConstruct(TNode* tnode, TNode* parent) {
	tnode->Process = NULL;
	nodeConstruct(tnodeGetNode(tnode));

	if(parent) {
		nodeAttach(tnodeGetNode(parent), tnodeGetNode(tnode));
	}

	tnodeReset(tnode);
}

inline void tnodeDestroy(TNode* tnode) {
	nodeDestroy(tnodeGetNode(tnode));
}

inline Node* tnodeGetNode(TNode* tnode) {
	return &tnode->node;
}

inline Matrix* tnodeGetMatrix(TNode* tnode) {
	return &tnode->mtx;
}

inline Matrix* tnodeGetTransformedMatrix(const TNode* node) {
	return node->resultMtx;
}

inline void tnodeReset(TNode* tnode) {
	identity_Matrix( tnodeGetMatrix( tnode ) );
}

void tnodeProcess(TNode* tnode, Matrix* resultMtx) {
	tnode->resultMtx = resultMtx;

	Node* node = tnodeGetNode(tnode);
	TNode* parent = (TNode*)nodeGetParent(node);

	if(parent) {
		multiply_Matrix(tnodeGetTransformedMatrix(tnode), tnodeGetMatrix(tnode), tnodeGetTransformedMatrix(parent));
	} else {
		memcpy(tnodeGetTransformedMatrix(tnode), tnodeGetMatrix(tnode), sizeof(Matrix));
	}

	Matrix* mtx = tnodeGetTransformedMatrix(tnode);
	tnode->csys.origin.x = MATRIX_GET(*mtx, 3, 0);
	tnode->csys.origin.y = MATRIX_GET(*mtx, 3, 1);
	tnode->csys.origin.z = MATRIX_GET(*mtx, 3, 2);
	tnode->csys.n.x = MATRIX_GET(*mtx, 2, 0);
	tnode->csys.n.y = MATRIX_GET(*mtx, 2, 1);
	tnode->csys.n.z = MATRIX_GET(*mtx, 2, 2);
}

inline void tnodeTransform(TNode* tnode, Matrix* mtx) {

	multiply_Matrix(tnodeGetMatrix(tnode), tnodeGetMatrix(tnode), mtx);
}

inline const Coordsys* tnodeGetCoordsys(const TNode* tnode) {

	return &tnode->csys;
}

