#include <stdlib/abort.h>
#include <actmgr/actmgr.h>
#include <containers/tree.h>
#include <engine.h>
#include <interpol/interpol.h>
#include <math/math.h>
#include <postproc/postproc.h>
#include <resmgr/resmgr.h>
#include <stdlib.h>
#include <sys/sys.h>
#include <sys/region.h>
#include <containers/vec_.h>
#include <string.h>
#include <malloc.h>
#include <ScalarField.h>
#include <ScalarFieldPolygonizer.h>
#include <mesh.h>

// BUG: display flickers if BILLBOARDS_NUM is too high (on PJ64); it doesn't if we don't iterate trough 2nd pass in engine.c
// Looks like the problem has something to do with length of the display list. On the other hand, CEN64 doesn't complain. Hm.

#define BILLBOARDS_NUM		 1
#define FLARE_TEXTURES_NUM	 6
//#define POSTPROC_BLUR

static Engine	engine;

typedef struct {
	Vector3f	pos;
	Vector3f	dir;
	float		angle;	/* degrees */
} Player;

typedef struct {
	Player	player;
	float 	angle;
} HandlerArg;

static void handle_vbi( void* sys_arg, void* user_arg ) {
	if( sys_arg ) {}

	if( NULL != user_arg ) {
		*((float*)user_arg) += 1.0f;	// angle
	}
}

static void handle_controller( void* sys_arg, void* user_arg ) {
	static float sty_old;
	static float stx_old;
	static float ud_old;

	if( (NULL != sys_arg) && (NULL != user_arg) ) {
		OSContPad const* rdata = (OSContPad const*)sys_arg;
		Player* player = (Player*)user_arg;

		player->angle -= 0.04f * stx_old + rdata->stick_x * 0.015f;

		player->dir.x = cosf( M_PI * (player->angle + 90) / 180 );
		//player->dir_y = player->dir.y;
		player->dir.z = -sinf( M_PI * (player->angle + 90) / 180 );

		player->pos.x += 0.0165 * (rdata->stick_y + sty_old) * player->dir.x;
		player->pos.z += 0.0165 * (rdata->stick_y + sty_old) * player->dir.z;
		player->pos.y += 0.03f * ud_old + (rdata->button & U_JPAD && TRUE);
		player->pos.y -= (rdata->button & D_JPAD && TRUE);

		sty_old = sty_old * 0.90f + 0.39f * rdata->stick_y;
		stx_old = 0.87f * stx_old + 0.012f * rdata->stick_x;
		ud_old = ud_old * 0.904f + (rdata->button & U_JPAD && TRUE) - (rdata->button & D_JPAD && TRUE);
	}
}

static float calculate_blobfield( float x, float y, float z, void const* arg ) {
	float g_d = *((float const*)arg);
	float x0 = x-g_d;
	float x1 = x+g_d;
	return 0.5f*( 1.0f/sqrtf( x0*x0 + y*y + z*z ) + 1.0f/sqrtf( x1*x1 + y*y + z*z ) );
}

static void process_task( Task const* task ) {
	GfxDevice* gdev = getGfxDevice_Sys();
	waitForGPU_GfxDevice( gdev );

	processTask_Engine( &engine, task );
	runGPU_GfxDevice( gdev, getDisplayList_Engine( &engine ) );
}

static void handle_img_action( Time const* time, void* arg ) {
	if( time ) {}
	if( arg ) {}

	static Raw_Xformer			root;
	static Camera_Xformer		camera_xf;
	static StaticMesh_Xformer	mesh_xf;
	static TreeNode				root_node, mesh_node, camera_node;
	static Task 				sceneTask = { (taskfptr)sceneTaskF_Engine, &root_node };

	static bool 				initialized = false;
	if( !initialized ) {
		// root
		root = createRaw_Engine( &engine );
		init_TreeNode( &root_node, &root );

		// camera
		camera_xf = createCamera_Engine( &engine, ORTHOGRAPHIC_CAMERA );
		/*
		setParams_OrthoCamera( getOrtho_Camera( camera_xf.camera ),
			-10.0f*SCREEN_ASPECT_RATIO, 10.0f*SCREEN_ASPECT_RATIO,
			-10.0f, 10.0f,
			1.33f, 1000.0f
		);
		*/
		setParams_OrthoCamera( getOrtho_Camera( camera_xf.camera ),
			-1.0f, 1.0f,
			-1.0f, 1.0f,
			-1.0f, 1.0f
		);
		init_TreeNode( &camera_node, &camera_xf );

		// mesh
		StaticMesh* mesh = (StaticMesh*)getResource_ResMgr( "img" );
		mesh_xf = createStaticMesh_Engine( &engine, mesh );
		init_TreeNode( &mesh_node, &mesh_xf );

		// setup hierarchy
		attachChild_TreeNode( &root_node, &mesh_node );
		attachChild_TreeNode( &root_node, &camera_node );

		initialized = true;
	}

	setCamera_Engine( &engine, &camera_xf );

	//float scale_factor = *((float*)arg);
	static Matrix mv;
	/*
	scale_Matrix( &mv,
		0.2f*sinf( scale_factor/10.0f ) + 2*sinf( scale_factor/100.0f ) + 4.0f,
		2*sinf( scale_factor/100.0f ) + 4.0f,
		1.0f
	);
	transformLocally_Xformer( camera_xf.xf, &mv );
	*/
	translate_Matrix( &mv,
		0.0f, 0.0f, -0.9999999f			// 0xF7EC
	);
	transformLocally_Xformer( mesh_xf.xf, &mv );

	//mesh_xf.static_mesh->opacity = 0.2f*sinf( scale_factor/100.0f ) + 0.5f;

	process_task( &sceneTask );

	// sample Z-buffer
	GfxDevice* gdev = getGfxDevice_Sys();
	waitForGPU_GfxDevice( gdev );

	Surface const* zbuf = getSurface_GfxDevice( gdev, SURFACE_Z );
	uint16_t zdev = *(uint16_t*)getValue_Surface( zbuf, 160, 100 );
	float zval = zAsFloat_GfxDevice( gdev, zdev );
	if( zval >= 0.999f ) {
		//abort();
	}
}

static void handle_scene_action( Time const* time, void* arg ) {
	static Material					materials[] = {
		MATERIAL_INIT(
			VECTOR4F_SET( 1.0f, 1.0f, 1.0f, 1.0f ),
			TEXTUREMAPPING_CONST,
			TEXTURE_ID_NONE
		)
	};
	static ScalarField				scalarField;
	static ScalarFieldPolygonizer	polygonizer;

	static Vector3f					randoms[BILLBOARDS_NUM];
	static Raw_Xformer				root;
	static Camera_Xformer			camera_xf;
	static StaticMesh_Xformer		worldmesh_xf,
									torusmesh_xf,
									torusmesh2_xf;
	static Mesh_Xformer				blobmesh_xf;
	static Lux_Xformer				lux_xf_r, lux_xf_b;
	static Billboard_Xformer		bb_xf[BILLBOARDS_NUM];
	static Flare_Xformer			flare_xf;
	static TreeNode					root_node,
									worldmesh_node,
									torusmesh_node,
									torusmesh2_node,
									blobmesh_node,
									lux_r_node,
									lux_b_node,
									camera_node,
									flare_node;
	static TreeNode					bb_node[BILLBOARDS_NUM];
	static Mesh						blobmesh;
	static Vec						flares;

	static bool initialized = false;
	if( !initialized ) {
		init_ScalarField( &scalarField, 9, 9, 9 );
		init_ScalarFieldPolygonizer( &polygonizer );

		addMaterial_Engine( &engine, &materials[0] );

		// root
		root = createRaw_Engine( &engine );
		init_TreeNode( &root_node, &root );

		// camera
		camera_xf = createCamera_Engine( &engine, PERSPECTIVE_CAMERA );
		setParams_PerspCamera( getPersp_Camera( camera_xf.camera ),
			63.74f, 1.33f, 1000.0f
		);
		init_TreeNode( &camera_node, &camera_xf );
		attachChild_TreeNode( &root_node, &camera_node );

		// meshes
		worldmesh_xf = createStaticMesh_Engine( &engine, (StaticMesh*)getResource_ResMgr( "world" ) );
		init_TreeNode( &worldmesh_node, &worldmesh_xf );
		attachChild_TreeNode( &root_node, &worldmesh_node );

		torusmesh_xf = createStaticMesh_Engine( &engine, (StaticMesh*)getResource_ResMgr( "torus" ) );	// large torus
		torusmesh_xf.static_mesh->opacity = 0.5f;
		init_TreeNode( &torusmesh_node, &torusmesh_xf );
		attachChild_TreeNode( &root_node, &torusmesh_node );

		torusmesh2_xf = createStaticMesh_Engine( &engine, (StaticMesh*)getResource_ResMgr( "torus" ) );	// small torus
		init_TreeNode( &torusmesh2_node, &torusmesh2_xf );
		attachChild_TreeNode( &torusmesh_node, &torusmesh2_node );

		blobmesh_xf = createMesh_Engine( &engine );
		init_TreeNode( &blobmesh_node, &blobmesh_xf );
		attachChild_TreeNode( &root_node, &blobmesh_node );

		// lights
		lux_xf_r = createLux_Engine( &engine );
		setColor_Lux( lux_xf_r.lux, 0.9f, 0.0f, 0.0f );
		init_TreeNode( &lux_r_node, &lux_xf_r );
		attachChild_TreeNode( &torusmesh2_node, &lux_r_node );

		lux_xf_b = createLux_Engine( &engine );
		setColor_Lux( lux_xf_b.lux, 0.0f, 0.0f, 0.9f );
		init_TreeNode( &lux_b_node, &lux_xf_b );
		attachChild_TreeNode( &root_node, &lux_b_node );

		// billboards
		for( uint32_t i=0; i<BILLBOARDS_NUM; i++ ) {
			randoms[i].x = 0.004f * rand() * (rand() > 0x4000 ? 1 : -1);
			randoms[i].y = 0.001f * rand() * (rand() > 0x4000 ? 1 : -1) + 50;
			randoms[i].z = 0.004f * rand() * (rand() > 0x4000 ? 1 : -1);

			bb_xf[i] = createBillboard_Engine( &engine, (StaticMesh*)getResource_ResMgr( "plane" ) );
			init_TreeNode( &bb_node[i], &bb_xf[i] );
			attachChild_TreeNode( &torusmesh2_node, &bb_node[i] );
		}

		// flares
		init_Vec( &flares, sizeof(Pair_StaticMesh_Xformer_Scale), ALIGN_DEFAULT );
		for( size_t i=0; i<FLARE_TEXTURES_NUM; ++i ) {
			static float scales[FLARE_TEXTURES_NUM] = { 0.01f, 0.005f, 0.004f, 0.007f, 0.008f, 0.009f };
			Pair_StaticMesh_Xformer_Scale pair = {
				.static_mesh_xf = createStaticMesh_Engine( &engine, (StaticMesh const*)getResource_ResMgr( "flare1" ) ),
				.scale = scales[i]
			};
			setTextureCombine_Material( &pair.static_mesh_xf.static_mesh->materials[0], TEXTURECOMBINE_ALPHA );
			pushBack_Vec( &flares, &pair );
		}
		flare_xf = createFlare_Engine( &engine, &flares );
		init_TreeNode( &flare_node, &flare_xf );
		attachChild_TreeNode( &torusmesh2_node, &flare_node );

		initialized = true;
	}

	static Matrix 		mv;
	static Task 		sceneTask = { (taskfptr)sceneTaskF_Engine, &root_node };
	static Task			flareTask = { (taskfptr)flareTaskF_Engine, &root_node };

	setCamera_Engine( &engine, &camera_xf );

	HandlerArg const* args = (HandlerArg*)arg;
	// camera
	rotate_Matrix( &mv, args->player.angle, 0, 1, 0 );
	transformLocally_Xformer( camera_xf.xf, &mv );
	translate_Matrix( &mv, args->player.pos.x, args->player.pos.y, args->player.pos.z );
	transformLocally_Xformer( camera_xf.xf, &mv );

	// flares
	translate_Matrix( &mv, 0, 0, 0 );
	transformLocally_Xformer( flare_xf.xf, &mv );
	
	// lights
	translate_Matrix( &mv, 0, 1, 0 );
	transformLocally_Xformer( lux_xf_r.xf, &mv );

	// torus
	rotate_Matrix( &mv, args->angle, 1, 1, 1 );
	transformLocally_Xformer( torusmesh_xf.xf, &mv );
	translate_Matrix( &mv, 0, 160, 0 );
	transformLocally_Xformer( torusmesh_xf.xf, &mv );

	// small torus
	scale_Matrix( &mv, 0.4f, 0.4f, 0.4f );
	transformLocally_Xformer( torusmesh2_xf.xf, &mv );
	Vector3f pos0 = VECTOR3F_SET( 0.0f, -30.0f, 0.0f );
	Vector3f pos1 = VECTOR3F_SET( 0.0f, 30.0f, 0.0f );
	Vector3f pos = calculate_LERP3f( &pos0, &pos1, sinf( 3.2 * time->msec / 1000.f ) );
	translate_Matrix( &mv, pos.x, pos.y, pos.z );
	transformLocally_Xformer( torusmesh2_xf.xf, &mv );

	// billboards
	for( uint32_t i=0; i<BILLBOARDS_NUM; i++ ) {
		translate_Matrix( &mv, randoms[i].x, randoms[i].y, -randoms[i].z );
		transformLocally_Xformer( bb_xf[i].xf, &mv );
	}

	GfxDevice* gdev = getGfxDevice_Sys();
	#ifdef POSTPROC_BLUR
		setRenderTarget_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_AUX ) );
	#else
		setRenderTarget_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_COLOR_DRAW ) );
	#endif

	// blobs
	float g_d = 0.4*sinf( args->angle / 100.0f );
	calculate_ScalarField( &scalarField, calculate_blobfield, &g_d );
	blobmesh = process_ScalarFieldPolygonizer( &polygonizer, &scalarField, 3.08f );
	*blobmesh_xf.mesh = blobmesh;
	translate_Matrix( &mv, 400, 160, 0 );
	transformLocally_Xformer( blobmesh_xf.xf, &mv );

	process_task( &sceneTask );
	process_task( &flareTask );

	destroy_Mesh( &blobmesh );
}

void mainproc( void* arg ) {
	if( arg ) {}

	HandlerArg function_args = {
		.player = {
			.pos = VECTOR3F_SET( 0.0f, 130.0f, 200.0f ),	// 400
			.dir = VECTOR3F_SET( 0.0f, 0.0f, 0.0f ),
			.angle = 0.0f	// 60
		},
		.angle = 0.0f
	};

	setAudioFrequency_Sys( 44100 / 2 );
	setAudioSource_Sys( (void*)((uintptr_t)getResource_ResMgr( "sound" ) + 44) );
	init_Engine( &engine );
	registerHandler_Sys( HANDLER_VBI, handle_vbi, &function_args.angle );
	registerHandler_Sys( HANDLER_CONTROLLER, handle_controller, &function_args );

	static Vector3f		clearColor = { 0.5f, 0.3f, 0.8f };
	static float		clearZ = 1.0f;
	static Task 		clearcfbTask = { (taskfptr)clearColorbufF_Engine, &clearColor };
	static Task 		clearzbufTask = { (taskfptr)clearZbufF_Engine, &clearZ };
	/* actions */
	ActMgr actmgr;
	init_ActMgr(&actmgr);

	Action sceneAction;
	Time tss = { 0, 1, 0 };
	Time tsd = ACTMGR_TIME_INFINITY;	//{ 0, 0, 0 };
	init_Action( &sceneAction, handle_scene_action, &function_args, tss, tsd );
	addAction_ActMgr( &actmgr, &sceneAction );

	Action imgAction;
	Time tid = { 20, 0, 0 };
	init_Action( &imgAction, handle_img_action, &function_args.angle, tss, tid );
	//addAction_ActMgr( &actmgr, &imgAction );

	GfxDevice* gdev = getGfxDevice_Sys();
	setZBuffer_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_Z ) );

	swapBuffers_GfxDevice( gdev );
	while(1) {
		#ifdef POSTPROC_BLUR
			setRenderTarget_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_AUX ) );
		#else
			setRenderTarget_Engine( &engine, getSurface_GfxDevice( gdev, SURFACE_COLOR_DRAW ) );
		#endif

		process_task( &clearcfbTask );
		process_task( &clearzbufTask );

		Time current_time = getTime_Sys();
		process_ActMgr( &actmgr, &current_time );
		waitForGPU_GfxDevice( gdev );
		#ifdef POSTPROC_BLUR
			Surface const* src = getSurface_GfxDevice( gdev, SURFACE_AUX );
			Surface const* dest = getSurface_GfxDevice( gdev, SURFACE_COLOR_DRAW );
			postprocBlur( dest, src );
		#endif
		swapBuffers_GfxDevice( gdev );
	}
}

